from lxml import etree

from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.lib.cib.tools import find_unique_id
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import RESOURCE_ROLES
from pcs.lib.xml_tools import export_attributes

ATTRIB = {
    "sequential": ("true", "false"),
    "require-all": ("true", "false"),
    "action": ("start", "promote", "demote", "stop"),
    "role": RESOURCE_ROLES,
}


def prepare_set(find_valid_id, resource_set):
    """return resource_set with corrected ids"""
    validate_options(resource_set["options"])
    return {
        "ids": [find_valid_id(id) for id in resource_set["ids"]],
        "options": resource_set["options"],
    }


def validate_options(options):
    # Pacemaker does not care currently about meaningfulness for concrete
    # constraint, so we use all attribs.
    for name, value in options.items():
        if name not in ATTRIB:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.InvalidOptions(
                        [name], sorted(ATTRIB.keys()), None
                    )
                )
            )
        if value not in ATTRIB[name]:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.InvalidOptionValue(
                        name, value, ATTRIB[name]
                    )
                )
            )


def create(parent, resource_set):
    """
    parent - lxml element for append new resource_set
    """
    element = etree.SubElement(parent, "resource_set")
    element.attrib.update(resource_set["options"])
    element.attrib["id"] = find_unique_id(
        parent.getroottree(),
        "{0}_set".format(parent.attrib.get("id", "constraint_set")),
    )

    for _id in resource_set["ids"]:
        etree.SubElement(element, "resource_ref").attrib["id"] = _id

    return element


def get_resource_id_set_list(element):
    return [
        resource_ref_element.attrib["id"]
        for resource_ref_element in element.findall(".//resource_ref")
    ]


def export(element):
    return {
        "ids": get_resource_id_set_list(element),
        "options": export_attributes(element),
    }



import pkgutil
from dataclasses import is_dataclass
from unittest import TestCase

import pcs
from pcs.common.interface.dto import DataTransferObject


def _import_all(_path):
    # arbitrary prefix so it doesn't iteract with real import in real tests
    for loader, module_name, is_pkg in pkgutil.walk_packages(
        _path, prefix="_pcs."
    ):
        if module_name.startswith("_pcs.snmp."):
            continue
        del is_pkg
        loader.find_module(module_name).load_module(module_name)


def _all_subclasses(cls):
    subclasses = set(cls.__subclasses__())
    return subclasses.union({s for c in subclasses for s in _all_subclasses(c)})


class DatatransferObjectTest(TestCase):
    def test_has_all_subclasses_are_dataclasses(self):
        _import_all(pcs.__path__)
        for cls in _all_subclasses(DataTransferObject):
            self.assertTrue(is_dataclass(cls), f"{cls} is not a dataclass")



from typing import (
    Any,
    Iterable,
    Dict,
    Type,
    TypeVar,
    Union,
)
from dataclasses import asdict, is_dataclass
import dacite

from pcs.common import types


PrimitiveType = Union[str, int, float, bool, None]
DtoPayload = Dict[str, "SerializableType"]  # type: ignore
SerializableType = Union[  # type: ignore
    PrimitiveType,
    DtoPayload,  # type: ignore
    Iterable["SerializableType"],  # type: ignore
]

T = TypeVar("T")


class DataTransferObject:
    pass


def to_dict(obj: DataTransferObject) -> DtoPayload:
    if not is_dataclass(obj):
        AssertionError()
    return asdict(obj)


DtoType = TypeVar("DtoType", bound=DataTransferObject)


def from_dict(cls: Type[DtoType], data: DtoPayload) -> DtoType:
    return dacite.from_dict(
        data_class=cls,
        data=data,
        # NOTE: all enum types has to be listed here in key cast
        # see: https://github.com/konradhalas/dacite#casting
        config=dacite.Config(
            cast=[
                types.CibNvsetType,
                types.CibRuleInEffectStatus,
                types.CibRuleExpressionType,
                types.DrRole,
                types.ResourceRelationType,
            ]
        ),
    )


class ImplementsToDto:
    def to_dto(self) -> Any:
        raise NotImplementedError()


class ImplementsFromDto:
    @classmethod
    def from_dto(cls: Type[T], dto_obj: Any) -> T:
        raise NotImplementedError()



import re
from textwrap import dedent
from unittest import TestCase

from pcs.cli import nvset
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
)
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.types import (
    CibNvsetType,
    CibRuleInEffectStatus,
    CibRuleExpressionType,
)


type_to_label = (
    (CibNvsetType.META, "Meta Attrs"),
    (CibNvsetType.INSTANCE, "Attributes"),
)


class NvsetDtoToLines(TestCase):
    @staticmethod
    def _export(dto, with_ids):
        return (
            "\n".join(nvset.nvset_dto_to_lines(dto, with_ids=with_ids)) + "\n"
        )

    def assert_lines(self, dto, lines):
        self.assertEqual(
            self._export(dto, True), lines,
        )
        self.assertEqual(
            self._export(dto, False), re.sub(r" +\(id:.*\)", "", lines),
        )

    def test_minimal(self):
        for nvtype, label in type_to_label:
            with self.subTest(nvset_type=nvtype, label=label):
                dto = CibNvsetDto("my-id", nvtype, {}, None, [])
                output = dedent(
                    f"""\
                      {label}: my-id
                    """
                )
                self.assert_lines(dto, output)

    def test_full(self):
        for nvtype, label in type_to_label:
            with self.subTest(nvset_type=nvtype, label=label):
                dto = CibNvsetDto(
                    "my-id",
                    nvtype,
                    {"score": "150"},
                    CibRuleExpressionDto(
                        "my-id-rule",
                        CibRuleExpressionType.RULE,
                        CibRuleInEffectStatus.UNKNOWN,
                        {"boolean-op": "or"},
                        None,
                        None,
                        [
                            CibRuleExpressionDto(
                                "my-id-rule-op",
                                CibRuleExpressionType.OP_EXPRESSION,
                                CibRuleInEffectStatus.UNKNOWN,
                                {"name": "monitor"},
                                None,
                                None,
                                [],
                                "op monitor",
                            ),
                        ],
                        "op monitor",
                    ),
                    [
                        CibNvpairDto("my-id-pair1", "name1", "value1"),
                        CibNvpairDto("my-id-pair2", "name 2", "value 2"),
                        CibNvpairDto("my-id-pair3", "name=3", "value=3"),
                    ],
                )
                output = dedent(
                    f"""\
                    {label}: my-id score=150
                      "name 2"="value 2"
                      name1=value1
                      "name=3"="value=3"
                      Rule: boolean-op=or (id:my-id-rule)
                        Expression: op monitor (id:my-id-rule-op)
                    """
                )
                self.assert_lines(dto, output)


class NvsetDtoListToLines(TestCase):
    @staticmethod
    def _export(dto, with_ids, include_expired):
        return (
            "\n".join(
                nvset.nvset_dto_list_to_lines(
                    dto, with_ids=with_ids, include_expired=include_expired
                )
            )
            + "\n"
        )

    def assert_lines(self, dto, include_expired, lines):
        self.assertEqual(
            self._export(dto, True, include_expired), lines,
        )
        self.assertEqual(
            self._export(dto, False, include_expired),
            re.sub(r" +\(id:.*\)", "", lines),
        )

    @staticmethod
    def fixture_dto(nvtype, in_effect):
        return CibNvsetDto(
            f"id-{in_effect}",
            nvtype,
            {"score": "150"},
            CibRuleExpressionDto(
                f"id-{in_effect}-rule",
                CibRuleExpressionType.RULE,
                in_effect,
                {"boolean-op": "or"},
                None,
                None,
                [
                    CibRuleExpressionDto(
                        f"id-{in_effect}-rule-op",
                        CibRuleExpressionType.OP_EXPRESSION,
                        CibRuleInEffectStatus.UNKNOWN,
                        {"name": "monitor"},
                        None,
                        None,
                        [],
                        "op monitor",
                    ),
                ],
                "op monitor",
            ),
            [CibNvpairDto(f"id-{in_effect}-pair1", "name1", "value1")],
        )

    def fixture_dto_list(self, nvtype):
        return [
            self.fixture_dto(nvtype, in_effect)
            for in_effect in CibRuleInEffectStatus
        ]

    def test_expired_included(self):
        self.maxDiff = None
        for nvtype, label in type_to_label:
            with self.subTest(nvset_type=nvtype, label=label):
                output = dedent(
                    f"""\
                    {label} (not yet in effect): id-NOT_YET_IN_EFFECT score=150
                      name1=value1
                      Rule (not yet in effect): boolean-op=or (id:id-NOT_YET_IN_EFFECT-rule)
                        Expression: op monitor (id:id-NOT_YET_IN_EFFECT-rule-op)
                    {label}: id-IN_EFFECT score=150
                      name1=value1
                      Rule: boolean-op=or (id:id-IN_EFFECT-rule)
                        Expression: op monitor (id:id-IN_EFFECT-rule-op)
                    {label} (expired): id-EXPIRED score=150
                      name1=value1
                      Rule (expired): boolean-op=or (id:id-EXPIRED-rule)
                        Expression: op monitor (id:id-EXPIRED-rule-op)
                    {label}: id-UNKNOWN score=150
                      name1=value1
                      Rule: boolean-op=or (id:id-UNKNOWN-rule)
                        Expression: op monitor (id:id-UNKNOWN-rule-op)
                """
                )
                self.assert_lines(self.fixture_dto_list(nvtype), True, output)

    def test_expired_excluded(self):
        self.maxDiff = None
        for nvtype, label in type_to_label:
            with self.subTest(nvset_type=nvtype, label=label):
                output = dedent(
                    f"""\
                    {label} (not yet in effect): id-NOT_YET_IN_EFFECT score=150
                      name1=value1
                      Rule (not yet in effect): boolean-op=or (id:id-NOT_YET_IN_EFFECT-rule)
                        Expression: op monitor (id:id-NOT_YET_IN_EFFECT-rule-op)
                    {label}: id-IN_EFFECT score=150
                      name1=value1
                      Rule: boolean-op=or (id:id-IN_EFFECT-rule)
                        Expression: op monitor (id:id-IN_EFFECT-rule-op)
                    {label}: id-UNKNOWN score=150
                      name1=value1
                      Rule: boolean-op=or (id:id-UNKNOWN-rule)
                        Expression: op monitor (id:id-UNKNOWN-rule-op)
                """
                )
                self.assert_lines(self.fixture_dto_list(nvtype), False, output)



from unittest import TestCase

from pcs.common.dr import DrRole
from pcs.lib.dr.config import facade


class Facade(TestCase):
    def test_create(self):
        for role in DrRole:
            with self.subTest(local_role=role.value):
                self.assertEqual(
                    dict(local=dict(role=role.value,), remote_sites=[],),
                    facade.Facade.create(role).config,
                )

    def test_local_role(self):
        for role in DrRole:
            with self.subTest(local_role=role.value):
                cfg = facade.Facade(
                    {"local": {"role": role.value,}, "remote_sites": [],}
                )
                self.assertEqual(cfg.local_role, role)

    def test_add_site(self):
        node_list = [f"node{i}" for i in range(4)]
        cfg = facade.Facade.create(DrRole.PRIMARY)
        cfg.add_site(DrRole.RECOVERY, node_list)
        self.assertEqual(
            dict(
                local=dict(role=DrRole.PRIMARY.value,),
                remote_sites=[
                    dict(
                        role=DrRole.RECOVERY.value,
                        nodes=[dict(name=node) for node in node_list],
                    ),
                ],
            ),
            cfg.config,
        )


class GetRemoteSiteList(TestCase):
    def test_no_sites(self):
        cfg = facade.Facade(
            {"local": {"role": DrRole.PRIMARY.value,}, "remote_sites": [],}
        )
        self.assertEqual(cfg.get_remote_site_list(), [])

    def test_one_site(self):
        cfg = facade.Facade(
            {
                "local": {"role": DrRole.PRIMARY.value,},
                "remote_sites": [
                    {
                        "role": DrRole.RECOVERY.value,
                        "nodes": [{"name": "node1"},],
                    },
                ],
            }
        )
        self.assertEqual(
            cfg.get_remote_site_list(),
            [facade.DrSite(role=DrRole.RECOVERY, node_name_list=["node1"]),],
        )

    def test_more_sites(self):
        cfg = facade.Facade(
            {
                "local": {"role": DrRole.RECOVERY.value,},
                "remote_sites": [
                    {
                        "role": DrRole.PRIMARY.value,
                        "nodes": [{"name": "nodeA1"}, {"name": "nodeA2"},],
                    },
                    {
                        "role": DrRole.RECOVERY.value,
                        "nodes": [{"name": "nodeB1"}, {"name": "nodeB2"},],
                    },
                ],
            }
        )
        self.assertEqual(
            cfg.get_remote_site_list(),
            [
                facade.DrSite(
                    role=DrRole.PRIMARY, node_name_list=["nodeA1", "nodeA2"]
                ),
                facade.DrSite(
                    role=DrRole.RECOVERY, node_name_list=["nodeB1", "nodeB2"]
                ),
            ],
        )

    def test_no_nodes(self):
        cfg = facade.Facade(
            {
                "local": {"role": DrRole.PRIMARY.value,},
                "remote_sites": [
                    {"role": DrRole.RECOVERY.value, "nodes": [],},
                ],
            }
        )
        self.assertEqual(
            cfg.get_remote_site_list(),
            [facade.DrSite(role=DrRole.RECOVERY, node_name_list=[]),],
        )



from pcs.lib.corosync import constants


class Section:
    def __init__(self, name):
        self._parent = None
        self._attr_list = []
        self._section_list = []
        self._name = str(name)

    @property
    def parent(self):
        return self._parent

    @property
    def name(self):
        return self._name

    @property
    def empty(self):
        return not self._attr_list and not self._section_list

    def export(self, indent="    "):
        lines = []
        for attr in self._attr_list:
            lines.append("{0}: {1}".format(*attr))
        if self._attr_list and self._section_list:
            lines.append("")
        section_count = len(self._section_list)
        for index, section in enumerate(self._section_list, 1):
            lines.extend(str(section).split("\n"))
            if not lines[-1].strip():
                del lines[-1]
            if index < section_count:
                lines.append("")
        if self.parent:
            lines = [indent + x if x else x for x in lines]
            lines.insert(0, self.name + " {")
            lines.append("}")
        final = "\n".join(lines)
        if final:
            final += "\n"
        return final

    def get_root(self):
        parent = self
        while parent.parent:
            parent = parent.parent
        return parent

    def get_attributes(self, name=None):
        return [
            attr for attr in self._attr_list if name is None or attr[0] == name
        ]

    def get_attributes_dict(self):
        return {attr[0]: attr[1] for attr in self._attr_list}

    def get_attribute_value(self, name, default=None):
        return self.get_attributes_dict().get(name, default)

    def add_attribute(self, name, value):
        name, value = str(name), str(value)
        self._attr_list.append([name, value])
        return self

    def del_attribute(self, attribute):
        self._attr_list = [
            attr for attr in self._attr_list if attr != attribute
        ]
        return self

    def del_attributes_by_name(self, name, value=None):
        self._attr_list = [
            attr
            for attr in self._attr_list
            if not (attr[0] == name and (value is None or attr[1] == value))
        ]
        return self

    def set_attribute(self, name, value):
        name, value = str(name), str(value)
        found = False
        new_attr_list = []
        for attr in self._attr_list:
            if attr[0] != name:
                new_attr_list.append(attr)
            elif not found:
                found = True
                attr[1] = value
                new_attr_list.append(attr)
        self._attr_list = new_attr_list
        if not found:
            self.add_attribute(name, value)
        return self

    def get_sections(self, name=None):
        return [
            section
            for section in self._section_list
            if name is None or section.name == name
        ]

    def add_section(self, section):
        parent = self
        while parent:
            if parent == section:
                raise CircularParentshipException()
            parent = parent.parent
        if section.parent:
            section.parent.del_section(section)
        # here we are editing obj's _parent attribute of the same class
        # pylint: disable=protected-access
        section._parent = self
        self._section_list.append(section)
        return self

    def del_section(self, section):
        self._section_list.remove(section)
        # don't set parent to None if the section was not found in the list
        # thanks to remove raising a ValueError in that case
        # here we are editing obj's _parent attribute of the same class
        # pylint: disable=protected-access
        section._parent = None
        return self

    def __str__(self):
        return self.export()


def parse_string(conf_text):
    root = Section("")
    _parse_section(conf_text.split("\n"), root)
    return root


def _parse_section(lines, section):
    # parser should work the same way as the original parser in corosync
    while lines:
        current_line = lines.pop(0).strip()
        if not current_line or current_line[0] == "#":
            continue
        if "{" in current_line:
            section_name_candidate, after_brace_junk = current_line.rsplit(
                "{", 1
            )
            if after_brace_junk.strip():
                raise ExtraCharactersAfterOpeningBraceException()
            section_name = section_name_candidate.strip()
            if not section_name:
                raise MissingSectionNameBeforeOpeningBraceException()
            new_section = Section(section_name.strip())
            section.add_section(new_section)
            _parse_section(lines, new_section)
        elif "}" in current_line:
            if current_line.strip() != "}":
                raise ExtraCharactersBeforeOrAfterClosingBraceException()
            if not section.parent:
                raise UnexpectedClosingBraceException()
            return
        elif ":" in current_line:
            section.add_attribute(
                *[x.strip() for x in current_line.split(":", 1)]
            )
        else:
            raise LineIsNotSectionNorKeyValueException()
    if section.parent:
        raise MissingClosingBraceException()


def verify_section(section, path_prefix=""):
    # prevents putting in any characters which break corosync.conf structure
    bad_section_name_list = []
    bad_attribute_name_list = []
    bad_attribute_value_list = []

    path = _prefix_path(path_prefix, section.name)
    # the root section has no name yet it is valid
    if section.name != "" and not _is_valid_name(section.name):
        bad_section_name_list.append(path)
    for name, value in section.get_attributes():
        if not _is_valid_name(name):
            bad_attribute_name_list.append(_prefix_path(path, name))
        if not _is_valid_value(value):
            bad_attribute_value_list.append((_prefix_path(path, name), value))
    for child_section in section.get_sections():
        bad_sections, bad_attr_names, bad_attr_values = verify_section(
            child_section, path
        )
        bad_section_name_list += bad_sections
        bad_attribute_name_list += bad_attr_names
        bad_attribute_value_list += bad_attr_values

    return (
        bad_section_name_list,
        bad_attribute_name_list,
        bad_attribute_value_list,
    )


def _prefix_path(prefix, path):
    return f"{prefix}.{path}" if prefix and path else path


def _is_valid_name(name):
    return constants.OPTION_NAME_RE.fullmatch(name) is not None


def _is_valid_value(value):
    # pylint: disable=superfluous-parens
    return not (set(value) & set("{}\n\r"))


class CorosyncConfParserException(Exception):
    pass


class CircularParentshipException(CorosyncConfParserException):
    pass


class ParseErrorException(CorosyncConfParserException):
    pass


class MissingClosingBraceException(ParseErrorException):
    pass


class UnexpectedClosingBraceException(ParseErrorException):
    pass


class MissingSectionNameBeforeOpeningBraceException(ParseErrorException):
    pass


class ExtraCharactersAfterOpeningBraceException(ParseErrorException):
    pass


class ExtraCharactersBeforeOrAfterClosingBraceException(ParseErrorException):
    pass


class LineIsNotSectionNorKeyValueException(ParseErrorException):
    pass



from pcs.common import file_type_codes
from pcs.lib.file import (
    metadata,
    raw_file,
    toolbox,
)


class FileInstance:
    """
    Provides a high-level easy access to config files
    """

    @classmethod
    def for_booth_config(cls, name, ghost_file=False, ghost_data=None):
        """
        Factory for booth config file

        string name -- booth instance name
        bool ghost_file -- use GhostFile if True, RealFile otherwise
        bytes ghost_data -- initial GhostFile data
        """
        return cls._for_booth(
            file_type_codes.BOOTH_CONFIG, name, ghost_file, ghost_data
        )

    @classmethod
    def for_booth_key(cls, name, ghost_file=False, ghost_data=None):
        """
        Factory for booth key file

        string name -- booth instance name
        bool ghost_file -- use GhostFile if True, RealFile otherwise
        bytes ghost_data -- initial GhostFile data
        """
        return cls._for_booth(
            file_type_codes.BOOTH_KEY, name, ghost_file, ghost_data
        )

    @classmethod
    def _for_booth(cls, file_type_code, name, ghost_file, ghost_data):
        return cls(
            _get_raw_file(
                metadata.for_file_type(file_type_code, name),
                ghost_file,
                ghost_data,
            ),
            toolbox.for_file_type(file_type_code),
        )

    @classmethod
    def for_known_hosts(cls):
        """
        Factory for known-hosts file
        """
        return cls._for_common(file_type_codes.PCS_KNOWN_HOSTS)

    @classmethod
    def for_pacemaker_key(cls):
        """
        Factory for pacemaker key file
        """
        return cls._for_common(file_type_codes.PACEMAKER_AUTHKEY)

    @classmethod
    def for_dr_config(cls) -> "FileInstance":
        """
        Factory for disaster-recovery config file
        """
        return cls._for_common(file_type_codes.PCS_DR_CONFIG)

    @classmethod
    def _for_common(
        cls, file_type_code: file_type_codes.FileTypeCode,
    ) -> "FileInstance":
        return cls(
            raw_file.RealFile(metadata.for_file_type(file_type_code)),
            toolbox.for_file_type(file_type_code),
        )

    def __init__(self, raw_file_interface, file_toolbox):
        """
        Factories should be used instead

        RawFileInterface raw_file_interface -- RealFile or GhostFile instance
        FileToolbox file_toolbox -- collection of file tools
        """
        self._raw_file = raw_file_interface
        self._toolbox = file_toolbox

    @property
    def raw_file(self):
        """
        Get the underlying RawFileInterface instance
        """
        return self._raw_file

    @property
    def toolbox(self):
        """
        Get the underlying FileToolbox instance
        """
        return self._toolbox

    def parser_exception_to_report_list(
        self, exception, force_code=None, is_forced_or_warning=False
    ):
        """
        Translate a RawFileError instance to a report

        RawFileError exception -- an exception to be translated
        string force_code -- is it a forcible error? by which code?
        bool is_forced_or_warning -- return a warning if True, error otherwise
        """
        return self._toolbox.parser.exception_to_report_list(
            exception,
            self._raw_file.metadata.file_type_code,
            (
                None
                if isinstance(self._raw_file, raw_file.GhostFile)
                else self._raw_file.metadata.path
            ),
            force_code,
            is_forced_or_warning,
        )

    def read_to_facade(self):
        return self._toolbox.facade(self.read_to_structure())

    def read_to_structure(self):
        return self._toolbox.parser.parse(self.read_raw())

    def read_raw(self):
        return self._raw_file.read()

    def raw_to_facade(self, raw_file_data):
        """
        Parse raw file data and return a corresponding facade

        bytes raw_file_data -- data to be parsed
        """
        return self._toolbox.facade(self.raw_to_structure(raw_file_data))

    def raw_to_structure(self, raw_file_data):
        """
        Parse raw file data and return a corresponding structure

        bytes raw_file_data -- data to be parsed
        """
        return self._toolbox.parser.parse(raw_file_data)

    def write_facade(self, facade, can_overwrite=False):
        self.write_structure(facade.config, can_overwrite=can_overwrite)

    def write_structure(self, structure, can_overwrite=False):
        self.write_raw(
            self._toolbox.exporter.export(structure),
            can_overwrite=can_overwrite,
        )

    def write_raw(self, raw_file_data, can_overwrite=False):
        """
        Write raw data to a file

        bytes raw_file_data -- data to be written
        bool can_overwrite -- raise if False and the file already exists
        """
        self._raw_file.write(raw_file_data, can_overwrite=can_overwrite)


def _get_raw_file(file_metadata, is_ghost, ghost_data):
    if is_ghost:
        return raw_file.GhostFile(file_metadata, file_data=ghost_data)
    return raw_file.RealFile(file_metadata)



import os.path

from pcs import utils

__pcs_location = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "pcs_for_tests"
)
# this can be changed from suite.py
test_installed = False


class PcsRunner:
    def __init__(self, cib_file, corosync_conf_opt=None, mock_settings=None):
        self.cib_file = cib_file
        self.corosync_conf_opt = corosync_conf_opt
        self.mock_settings = mock_settings

    def run(self, args):
        return pcs(
            self.cib_file,
            args,
            corosync_conf_opt=self.corosync_conf_opt,
            mock_settings=self.mock_settings,
        )


def pcs(cib_file, args, corosync_conf_opt=None, mock_settings=None):
    """
    Run pcs with -f on specified file
    Return tuple with:
        shell stdoutdata
        shell returncode
    """
    if mock_settings is None:
        mock_settings = {}

    env_mock_settings_prefix = "PCS.SETTINGS."
    env = {
        "{}{}".format(env_mock_settings_prefix, option): value
        for option, value in mock_settings.items()
    }
    if test_installed:
        env["PCS_TEST.TEST_INSTALLED"] = "1"

    cmd = [__pcs_location] + args
    if cib_file:
        cmd.extend(["-f", cib_file])
    if corosync_conf_opt:
        cmd.extend(["--corosync_conf", corosync_conf_opt])

    return utils.run(cmd, env_extend=env)



from .types import (
    BoothConfigUsedWhere,
    DefaultAddressSource,
    FenceHistoryCommandType,
    PcsCommand,
    ReasonType,
    ServiceAction,
)


BOOTH_CONFIG_USED_IN_CLUSTER_RESOURCE = BoothConfigUsedWhere(
    "in a cluster resource"
)
BOOTH_CONFIG_USED_ENABLED_IN_SYSTEMD = BoothConfigUsedWhere(
    "enabled in systemd"
)
BOOTH_CONFIG_USED_RUNNING_IN_SYSTEMD = BoothConfigUsedWhere(
    "running in systemd"
)

FENCE_HISTORY_COMMAND_CLEANUP = FenceHistoryCommandType("cleanup")
FENCE_HISTORY_COMMAND_SHOW = FenceHistoryCommandType("show")
FENCE_HISTORY_COMMAND_UPDATE = FenceHistoryCommandType("update")

PCS_COMMAND_OPERATION_DEFAULTS_UPDATE = PcsCommand(
    "resource op defaults update"
)
PCS_COMMAND_RESOURCE_DEFAULTS_UPDATE = PcsCommand("resource defaults update")

SERVICE_ACTION_START = ServiceAction("START")
SERVICE_ACTION_STOP = ServiceAction("STOP")
SERVICE_ACTION_ENABLE = ServiceAction("ENABLE")
SERVICE_ACTION_DISABLE = ServiceAction("DISABLE")
SERVICE_ACTION_KILL = ServiceAction("KILL")

REASON_UNREACHABLE = ReasonType("unreachable")
REASON_NOT_LIVE_CIB = ReasonType("not_live_cib")

DEFAULT_ADDRESS_SOURCE_KNOWN_HOSTS = DefaultAddressSource("known_hosts")
DEFAULT_ADDRESS_SOURCE_HOST_NAME = DefaultAddressSource("host_name")



from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.bin_mock import get_mock_settings
from pcs_test.tools.misc import (
    get_test_resource as rc,
    get_tmp_file,
    outdent,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


class OldCibPushTest(TestCase, AssertPcsMixin):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_misc")
        write_file_to_tmpfile(rc("cib-empty-1.2.xml"), self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.pcs_runner.mock_settings = get_mock_settings("crm_resource_binary")

    def tearDown(self):
        self.temp_cib.close()

    def test_warning_old_push(self):
        self.assert_pcs_success(
            "resource create dummy ocf:pacemaker:Dummy --no-default-ops".split(),
            "Warning: Replacing the whole CIB instead of applying a diff, "
            "a race condition may happen if the CIB is pushed more than "
            "once simultaneously. To fix this, upgrade pacemaker to get "
            "crm_feature_set at least 3.0.9, current is 3.0.8.\n",
        )
        self.assert_pcs_success(
            "resource config".split(),
            # pylint: disable=line-too-long
            # fmt: off
            outdent(
            """\
             Resource: dummy (class=ocf provider=pacemaker type=Dummy)
              Operations: monitor interval=10s timeout=20s (dummy-monitor-interval-10s)
            """
            ),
            # fmt: on
        )



# pylint: disable=too-many-lines
from contextlib import contextmanager
from functools import partial
from typing import (
    cast,
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
)

from lxml.etree import _Element

from pcs.common import file_type_codes
from pcs.common.interface import dto
from pcs.common import reports
from pcs.common.reports import ReportItemList
from pcs.common.reports.item import ReportItem
from pcs.common.tools import Version
from pcs.lib.cib import (
    resource,
    status as cib_status,
)
from pcs.lib.cib.tag import (
    expand_tag,
    TAG_TAG,
)
from pcs.lib.cib.tools import (
    find_element_by_tag_and_id,
    get_resources,
    get_status,
    IdProvider,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.external import CommandRunner
from pcs.lib.node import get_existing_nodes_names_addrs
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker import simulate as simulate_tools
from pcs.lib.pacemaker.live import (
    has_resource_unmove_unban_expired_support,
    resource_ban,
    resource_move,
    resource_unmove_unban,
    simulate_cib,
    wait_for_idle,
)
from pcs.lib.pacemaker.state import (
    ensure_resource_state,
    get_resource_state,
    info_resource_state,
    is_resource_managed,
    ResourceNotFound,
)
from pcs.lib.pacemaker.values import (
    timeout_to_seconds,
    validate_id,
)
from pcs.lib.resource_agent import (
    find_valid_resource_agent_by_name as get_agent,
)
from pcs.lib.validate import ValueTimeInterval


WaitType = Union[None, bool, int]


@contextmanager
def resource_environment(
    env,
    wait=False,
    wait_for_resource_ids=None,
    resource_state_reporter=info_resource_state,
    required_cib_version=None,
):
    env.ensure_wait_satisfiable(wait)
    yield get_resources(env.get_cib(required_cib_version))
    _push_cib_wait(env, wait, wait_for_resource_ids, resource_state_reporter)


def _push_cib_wait(
    env: LibraryEnvironment,
    wait: WaitType = False,
    wait_for_resource_ids: Optional[Iterable[str]] = None,
    resource_state_reporter: Callable[
        [_Element, str], ReportItem
    ] = info_resource_state,
) -> None:
    env.push_cib(wait=wait)
    if wait is not False and wait_for_resource_ids:
        state = env.get_cluster_state()
        if env.report_processor.report_list(
            [
                resource_state_reporter(state, res_id)
                for res_id in wait_for_resource_ids
            ]
        ).has_errors:
            raise LibraryError()


def _ensure_disabled_after_wait(disabled_after_wait):
    def inner(state, resource_id):
        return ensure_resource_state(
            not disabled_after_wait, state, resource_id
        )

    return inner


def _validate_remote_connection(
    resource_agent,
    existing_nodes_addrs,
    resource_id,
    instance_attributes,
    allow_not_suitable_command,
):
    if resource_agent.get_name() != resource.remote_node.AGENT_NAME.full_name:
        return []

    report_list = []
    report_list.append(
        ReportItem(
            severity=reports.item.get_severity(
                reports.codes.FORCE_NOT_SUITABLE_COMMAND,
                allow_not_suitable_command,
            ),
            message=reports.messages.UseCommandNodeAddRemote(),
        )
    )

    report_list.extend(
        resource.remote_node.validate_host_not_conflicts(
            existing_nodes_addrs, resource_id, instance_attributes
        )
    )
    return report_list


def _validate_guest_change(
    tree,
    existing_nodes_names,
    existing_nodes_addrs,
    meta_attributes,
    allow_not_suitable_command,
    detect_remove=False,
):
    if not resource.guest_node.is_node_name_in_options(meta_attributes):
        return []

    node_name = resource.guest_node.get_node_name_from_options(meta_attributes)

    report_list = []
    if detect_remove and not resource.guest_node.get_guest_option_value(
        meta_attributes
    ):
        report_msg = reports.messages.UseCommandNodeRemoveGuest()
    else:
        report_msg = reports.messages.UseCommandNodeAddGuest()

    report_list.append(
        ReportItem(
            severity=reports.item.get_severity(
                reports.codes.FORCE_NOT_SUITABLE_COMMAND,
                allow_not_suitable_command,
            ),
            message=report_msg,
        )
    )

    report_list.extend(
        resource.guest_node.validate_conflicts(
            tree,
            existing_nodes_names,
            existing_nodes_addrs,
            node_name,
            meta_attributes,
        )
    )

    return report_list


def _get_nodes_to_validate_against(env, tree):
    if not env.is_corosync_conf_live and env.is_cib_live:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentRequired(
                    [file_type_codes.COROSYNC_CONF]
                )
            )
        )

    if not env.is_cib_live and env.is_corosync_conf_live:
        # we do not try to get corosync.conf from live cluster when cib is not
        # taken from live cluster
        return get_existing_nodes_names_addrs(cib=tree)

    return get_existing_nodes_names_addrs(env.get_corosync_conf(), cib=tree)


def _check_special_cases(
    env,
    resource_agent,
    resources_section,
    resource_id,
    meta_attributes,
    instance_attributes,
    allow_not_suitable_command,
):
    # fmt: off
    if (
        resource_agent.get_name() != resource.remote_node.AGENT_NAME.full_name
        and
        not resource.guest_node.is_node_name_in_options(meta_attributes)
    ):
        # if no special case happens we won't take care about corosync.conf that
        # is needed for getting nodes to validate against
        return
    # fmt: on

    (
        existing_nodes_names,
        existing_nodes_addrs,
        report_list,
    ) = _get_nodes_to_validate_against(env, resources_section)

    report_list.extend(
        _validate_remote_connection(
            resource_agent,
            existing_nodes_addrs,
            resource_id,
            instance_attributes,
            allow_not_suitable_command,
        )
    )
    report_list.extend(
        _validate_guest_change(
            resources_section,
            existing_nodes_names,
            existing_nodes_addrs,
            meta_attributes,
            allow_not_suitable_command,
        )
    )

    if env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()


_find_bundle = partial(find_element_by_tag_and_id, resource.bundle.TAG)


def _get_required_cib_version_for_container(
    container_options, container_type=None
):
    if container_type == "podman":
        return Version(3, 2, 0)

    if "promoted-max" in container_options:
        return Version(3, 0, 0)

    if container_type == "rkt":
        return Version(2, 10, 0)

    return Version(2, 8, 0)


def create(
    env: LibraryEnvironment,
    resource_id: str,
    resource_agent_name: str,
    operation_list: Iterable[Mapping[str, str]],
    meta_attributes: Mapping[str, str],
    instance_attributes: Mapping[str, str],
    allow_absent_agent: bool = False,
    allow_invalid_operation: bool = False,
    allow_invalid_instance_attributes: bool = False,
    use_default_operations: bool = True,
    ensure_disabled: bool = False,
    wait: WaitType = False,
    allow_not_suitable_command: bool = False,
):
    # pylint: disable=too-many-arguments, too-many-locals
    """
    Create a primitive resource in a cib.

    env -- provides all for communication with externals
    resource_id -- is identifier of resource
    resource_agent_name -- contains name for the identification of agent
    operation_list -- contains attributes for each entered operation
    meta_attributes -- contains attributes for primitive/meta_attributes
    instance_attributes -- contains attributes for primitive/instance_attributes
    allow_absent_agent -- is a flag for allowing agent that is not installed
        in a system
    allow_invalid_operation -- is a flag for allowing to use operations that
        are not listed in a resource agent metadata
    allow_invalid_instance_attributes -- is a flag for allowing to use
        instance attributes that are not listed in a resource agent metadata
        or for allowing to not use the instance_attributes that are required in
        resource agent metadata
    use_default_operations -- is a flag for stopping stopping of adding
        default cib operations (specified in a resource agent)
    ensure_disabled -- is flag that keeps resource in target-role "Stopped"
    wait -- is flag for controlling waiting for pacemaker idle mechanism
    allow_not_suitable_command -- flag for FORCE_NOT_SUITABLE_COMMAND
    """
    resource_agent = get_agent(
        env.report_processor,
        env.cmd_runner(),
        resource_agent_name,
        allow_absent_agent,
    )
    with resource_environment(
        env,
        wait,
        [resource_id],
        _ensure_disabled_after_wait(
            ensure_disabled
            or resource.common.are_meta_disabled(meta_attributes)
        ),
        required_cib_version=get_required_cib_version_for_primitive(
            operation_list
        ),
    ) as resources_section:
        id_provider = IdProvider(resources_section)
        _check_special_cases(
            env,
            resource_agent,
            resources_section,
            resource_id,
            meta_attributes,
            instance_attributes,
            allow_not_suitable_command,
        )

        primitive_element = resource.primitive.create(
            env.report_processor,
            resources_section,
            id_provider,
            resource_id,
            resource_agent,
            operation_list,
            meta_attributes,
            instance_attributes,
            allow_invalid_operation,
            allow_invalid_instance_attributes,
            use_default_operations,
        )
        if ensure_disabled:
            resource.common.disable(primitive_element, id_provider)


def create_as_clone(
    env: LibraryEnvironment,
    resource_id: str,
    resource_agent_name: str,
    operation_list: Iterable[Mapping[str, str]],
    meta_attributes: Mapping[str, str],
    instance_attributes: Mapping[str, str],
    clone_meta_options: Mapping[str, str],
    clone_id: Optional[str] = None,
    allow_absent_agent: bool = False,
    allow_invalid_operation: bool = False,
    allow_invalid_instance_attributes: bool = False,
    use_default_operations: bool = True,
    ensure_disabled: bool = False,
    wait: WaitType = False,
    allow_not_suitable_command: bool = False,
):
    # pylint: disable=too-many-arguments, too-many-locals
    """
    Create a primitive resource in a clone

    env -- provides all for communication with externals
    resource_id -- is identifier of resource
    resource_agent_name -- contains name for the identification of agent
    operation_list -- contains attributes for each entered operation
    meta_attributes -- contains attributes for primitive/meta_attributes
    instance_attributes -- contains attributes for primitive/instance_attributes
    clone_meta_options -- contains attributes for clone/meta_attributes
    clone_id -- optional custom clone id, if not set then clone id is generated
        from primitive resource or group id
    allow_absent_agent -- is a flag for allowing agent that is not installed
        in a system
    allow_invalid_operation -- is a flag for allowing to use operations that
        are not listed in a resource agent metadata
    allow_invalid_instance_attributes -- is a flag for allowing to use
        instance attributes that are not listed in a resource agent metadata
        or for allowing to not use the instance_attributes that are required in
        resource agent metadata
    use_default_operations -- is a flag for stopping stopping of adding
        default cib operations (specified in a resource agent)
    ensure_disabled -- is flag that keeps resource in target-role "Stopped"
    wait -- is flag for controlling waiting for pacemaker idle mechanism
    allow_not_suitable_command -- flag for FORCE_NOT_SUITABLE_COMMAND
    """
    resource_agent = get_agent(
        env.report_processor,
        env.cmd_runner(),
        resource_agent_name,
        allow_absent_agent,
    )
    with resource_environment(
        env,
        wait,
        [resource_id],
        _ensure_disabled_after_wait(
            ensure_disabled
            or resource.common.are_meta_disabled(meta_attributes)
            or resource.common.is_clone_deactivated_by_meta(clone_meta_options)
        ),
        required_cib_version=get_required_cib_version_for_primitive(
            operation_list
        ),
    ) as resources_section:
        id_provider = IdProvider(resources_section)
        _check_special_cases(
            env,
            resource_agent,
            resources_section,
            resource_id,
            meta_attributes,
            instance_attributes,
            allow_not_suitable_command,
        )

        primitive_element = resource.primitive.create(
            env.report_processor,
            resources_section,
            id_provider,
            resource_id,
            resource_agent,
            operation_list,
            meta_attributes,
            instance_attributes,
            allow_invalid_operation,
            allow_invalid_instance_attributes,
            use_default_operations,
        )
        if clone_id is not None:
            if env.report_processor.report_list(
                resource.clone.validate_clone_id(clone_id, id_provider),
            ).has_errors:
                raise LibraryError()
        clone_element = resource.clone.append_new(
            resources_section,
            id_provider,
            primitive_element,
            clone_meta_options,
            clone_id=clone_id,
        )
        if ensure_disabled:
            resource.common.disable(clone_element, id_provider)


def create_in_group(
    env: LibraryEnvironment,
    resource_id: str,
    resource_agent_name: str,
    group_id: str,
    operation_list: Iterable[Mapping[str, str]],
    meta_attributes: Mapping[str, str],
    instance_attributes: Mapping[str, str],
    allow_absent_agent: bool = False,
    allow_invalid_operation: bool = False,
    allow_invalid_instance_attributes: bool = False,
    use_default_operations: bool = True,
    ensure_disabled: bool = False,
    adjacent_resource_id: Optional[str] = None,
    put_after_adjacent: bool = False,
    wait: WaitType = False,
    allow_not_suitable_command: bool = False,
):
    # pylint: disable=too-many-arguments, too-many-locals
    """
    Create resource in a cib and put it into defined group

    env -- provides all for communication with externals
    resource_id -- is identifier of resource
    resource_agent_name -- contains name for the identification of agent
    group_id -- is identificator for group to put primitive resource inside
    operation_list -- contains attributes for each entered operation
    meta_attributes -- contains attributes for primitive/meta_attributes
    instance_attributes -- contains attributes for primitive/instance_attributes
    allow_absent_agent -- is a flag for allowing agent that is not installed
        in a system
    allow_invalid_operation -- is a flag for allowing to use operations that
        are not listed in a resource agent metadata
    allow_invalid_instance_attributes -- is a flag for allowing to use
        instance attributes that are not listed in a resource agent metadata
        or for allowing to not use the instance_attributes that are required in
        resource agent metadata
    use_default_operations -- is a flag for stopping stopping of adding
        default cib operations (specified in a resource agent)
    ensure_disabled -- is flag that keeps resource in target-role "Stopped"
    adjacent_resource_id -- identify neighbor of a newly created resource
    put_after_adjacent -- is flag to put a newly create resource befor/after
        adjacent resource
    wait -- is flag for controlling waiting for pacemaker idle mechanism
    allow_not_suitable_command -- flag for FORCE_NOT_SUITABLE_COMMAND
    """
    resource_agent = get_agent(
        env.report_processor,
        env.cmd_runner(),
        resource_agent_name,
        allow_absent_agent,
    )
    with resource_environment(
        env,
        wait,
        [resource_id],
        _ensure_disabled_after_wait(
            ensure_disabled
            or resource.common.are_meta_disabled(meta_attributes)
        ),
        required_cib_version=get_required_cib_version_for_primitive(
            operation_list
        ),
    ) as resources_section:
        id_provider = IdProvider(resources_section)
        _check_special_cases(
            env,
            resource_agent,
            resources_section,
            resource_id,
            meta_attributes,
            instance_attributes,
            allow_not_suitable_command,
        )

        primitive_element = resource.primitive.create(
            env.report_processor,
            resources_section,
            id_provider,
            resource_id,
            resource_agent,
            operation_list,
            meta_attributes,
            instance_attributes,
            allow_invalid_operation,
            allow_invalid_instance_attributes,
            use_default_operations,
        )
        if ensure_disabled:
            resource.common.disable(primitive_element, id_provider)
        validate_id(group_id, "group name")
        resource.group.place_resource(
            resource.group.provide_group(resources_section, group_id),
            primitive_element,
            adjacent_resource_id,
            put_after_adjacent,
        )


def create_into_bundle(
    env: LibraryEnvironment,
    resource_id: str,
    resource_agent_name: str,
    operation_list: Iterable[Mapping[str, str]],
    meta_attributes: Mapping[str, str],
    instance_attributes: Mapping[str, str],
    bundle_id: str,
    allow_absent_agent: bool = False,
    allow_invalid_operation: bool = False,
    allow_invalid_instance_attributes: bool = False,
    use_default_operations: bool = True,
    ensure_disabled: bool = False,
    wait: WaitType = False,
    allow_not_suitable_command: bool = False,
    allow_not_accessible_resource: bool = False,
):
    # pylint: disable=too-many-arguments, too-many-locals
    """
    Create a new resource in a cib and put it into an existing bundle

    env -- provides all for communication with externals
    resource_id -- is identifier of resource
    resource_agent_name -- contains name for the identification of agent
    operation_list -- contains attributes for each entered operation
    meta_attributes -- contains attributes for primitive/meta_attributes
    instance_attributes -- contains attributes for
        primitive/instance_attributes
    bundle_id -- is id of an existing bundle to put the created resource in
    allow_absent_agent -- is a flag for allowing agent that is not installed
        in a system
    allow_invalid_operation -- is a flag for allowing to use operations that
        are not listed in a resource agent metadata
    allow_invalid_instance_attributes -- is a flag for allowing to use
        instance attributes that are not listed in a resource agent metadata
        or for allowing to not use the instance_attributes that are required in
        resource agent metadata
    use_default_operations -- is a flag for stopping stopping of adding
        default cib operations (specified in a resource agent)
    ensure_disabled -- is flag that keeps resource in target-role "Stopped"
    wait -- is flag for controlling waiting for pacemaker idle mechanism
    allow_not_suitable_command -- flag for FORCE_NOT_SUITABLE_COMMAND
    allow_not_accessible_resource -- flag for
        FORCE_RESOURCE_IN_BUNDLE_NOT_ACCESSIBLE
    """
    resource_agent = get_agent(
        env.report_processor,
        env.cmd_runner(),
        resource_agent_name,
        allow_absent_agent,
    )
    required_cib_version = get_required_cib_version_for_primitive(
        operation_list
    )
    if not required_cib_version:
        required_cib_version = Version(2, 8, 0)
    with resource_environment(
        env,
        wait,
        [resource_id],
        _ensure_disabled_after_wait(
            ensure_disabled
            or resource.common.are_meta_disabled(meta_attributes)
        ),
        required_cib_version=required_cib_version,
    ) as resources_section:
        id_provider = IdProvider(resources_section)
        _check_special_cases(
            env,
            resource_agent,
            resources_section,
            resource_id,
            meta_attributes,
            instance_attributes,
            allow_not_suitable_command,
        )

        primitive_element = resource.primitive.create(
            env.report_processor,
            resources_section,
            id_provider,
            resource_id,
            resource_agent,
            operation_list,
            meta_attributes,
            instance_attributes,
            allow_invalid_operation,
            allow_invalid_instance_attributes,
            use_default_operations,
        )
        if ensure_disabled:
            resource.common.disable(primitive_element, id_provider)

        bundle_el = _find_bundle(resources_section, bundle_id)
        if not resource.bundle.is_pcmk_remote_accessible(bundle_el):
            if env.report_processor.report(
                ReportItem(
                    severity=reports.item.get_severity(
                        reports.codes.FORCE_RESOURCE_IN_BUNDLE_NOT_ACCESSIBLE,
                        allow_not_accessible_resource,
                    ),
                    message=reports.messages.ResourceInBundleNotAccessible(
                        bundle_id, resource_id,
                    ),
                )
            ).has_errors:
                raise LibraryError()
        resource.bundle.add_resource(bundle_el, primitive_element)


def bundle_create(
    env,
    bundle_id,
    container_type,
    container_options=None,
    network_options=None,
    port_map=None,
    storage_map=None,
    meta_attributes=None,
    force_options=False,
    ensure_disabled=False,
    wait=False,
):
    # pylint: disable=too-many-arguments
    """
    Create a new bundle containing no resources

    LibraryEnvironment env -- provides communication with externals
    string bundle_id -- id of the new bundle
    string container_type -- container engine name (docker, lxc...)
    dict container_options -- container options
    dict network_options -- network options
    list of dict port_map -- a list of port mapping options
    list of dict storage_map -- a list of storage mapping options
    dict meta_attributes -- bundle's meta attributes
    bool force_options -- return warnings instead of forceable errors
    bool ensure_disabled -- set the bundle's target-role to "Stopped"
    mixed wait -- False: no wait, None: wait default timeout, int: wait timeout
    """
    container_options = container_options or {}
    network_options = network_options or {}
    port_map = port_map or []
    storage_map = storage_map or []
    meta_attributes = meta_attributes or {}

    with resource_environment(
        env,
        wait,
        [bundle_id],
        _ensure_disabled_after_wait(
            ensure_disabled
            or resource.common.are_meta_disabled(meta_attributes)
        ),
        required_cib_version=_get_required_cib_version_for_container(
            container_options, container_type,
        ),
    ) as resources_section:
        # no need to run validations related to remote and guest nodes as those
        # nodes can only be created from primitive resources
        id_provider = IdProvider(resources_section)
        if env.report_processor.report_list(
            resource.bundle.validate_new(
                id_provider,
                bundle_id,
                container_type,
                container_options,
                network_options,
                port_map,
                storage_map,
                # TODO meta attributes - there is no validation for now
                force_options,
            )
        ).has_errors:
            raise LibraryError()
        bundle_element = resource.bundle.append_new(
            resources_section,
            id_provider,
            bundle_id,
            container_type,
            container_options,
            network_options,
            port_map,
            storage_map,
            meta_attributes,
        )
        if ensure_disabled:
            resource.common.disable(bundle_element, id_provider)


def bundle_reset(
    env,
    bundle_id,
    container_options=None,
    network_options=None,
    port_map=None,
    storage_map=None,
    meta_attributes=None,
    force_options=False,
    ensure_disabled=False,
    wait=False,
):
    # pylint: disable=too-many-arguments
    """
    Remove configuration of bundle bundle_id and create new one into it.

    LibraryEnvironment env -- provides communication with externals
    string bundle_id -- id of the bundle to reset
    dict container_options -- container options
    dict network_options -- network options
    list of dict port_map -- a list of port mapping options
    list of dict storage_map -- a list of storage mapping options
    dict meta_attributes -- bundle's meta attributes
    bool force_options -- return warnings instead of forceable errors
    bool ensure_disabled -- set the bundle's target-role to "Stopped"
    mixed wait -- False: no wait, None: wait default timeout, int: wait timeout
    """
    container_options = container_options or {}
    network_options = network_options or {}
    port_map = port_map or []
    storage_map = storage_map or []
    meta_attributes = meta_attributes or {}

    with resource_environment(
        env,
        wait,
        [bundle_id],
        _ensure_disabled_after_wait(
            ensure_disabled
            or resource.common.are_meta_disabled(meta_attributes)
        ),
        required_cib_version=_get_required_cib_version_for_container(
            container_options
        ),
    ) as resources_section:
        bundle_element = _find_bundle(resources_section, bundle_id)
        if env.report_processor.report_list(
            resource.bundle.validate_reset_to_minimal(bundle_element)
        ).has_errors:
            raise LibraryError()
        resource.bundle.reset_to_minimal(bundle_element)

        id_provider = IdProvider(resources_section)
        if env.report_processor.report_list(
            resource.bundle.validate_reset(
                id_provider,
                bundle_element,
                container_options,
                network_options,
                port_map,
                storage_map,
                # TODO meta attributes - there is no validation for now
                force_options,
            )
        ).has_errors:
            raise LibraryError()

        resource.bundle.update(
            id_provider,
            bundle_element,
            container_options,
            network_options,
            port_map_add=port_map,
            port_map_remove=[],
            storage_map_add=storage_map,
            storage_map_remove=[],
            meta_attributes=meta_attributes,
        )

        if ensure_disabled:
            resource.common.disable(bundle_element, id_provider)


def bundle_update(
    env,
    bundle_id,
    container_options=None,
    network_options=None,
    port_map_add=None,
    port_map_remove=None,
    storage_map_add=None,
    storage_map_remove=None,
    meta_attributes=None,
    force_options=False,
    wait=False,
):
    # pylint: disable=too-many-arguments
    """
    Modify an existing bundle (does not touch encapsulated resources)

    LibraryEnvironment env -- provides communication with externals
    string bundle_id -- id of the bundle to modify
    dict container_options -- container options to modify
    dict network_options -- network options to modify
    list of dict port_map_add -- list of port mapping options to add
    list of string port_map_remove -- list of port mapping ids to remove
    list of dict storage_map_add -- list of storage mapping options to add
    list of string storage_map_remove -- list of storage mapping ids to remove
    dict meta_attributes -- meta attributes to update
    bool force_options -- return warnings instead of forceable errors
    mixed wait -- False: no wait, None: wait default timeout, int: wait timeout
    """
    container_options = container_options or {}
    network_options = network_options or {}
    port_map_add = port_map_add or []
    port_map_remove = port_map_remove or []
    storage_map_add = storage_map_add or []
    storage_map_remove = storage_map_remove or []
    meta_attributes = meta_attributes or {}

    with resource_environment(
        env,
        wait,
        [bundle_id],
        required_cib_version=_get_required_cib_version_for_container(
            container_options
        ),
    ) as resources_section:
        # no need to run validations related to remote and guest nodes as those
        # nodes can only be created from primitive resources
        id_provider = IdProvider(resources_section)
        bundle_element = _find_bundle(resources_section, bundle_id)
        if env.report_processor.report_list(
            resource.bundle.validate_update(
                id_provider,
                bundle_element,
                container_options,
                network_options,
                port_map_add,
                port_map_remove,
                storage_map_add,
                storage_map_remove,
                # TODO meta attributes - there is no validation for now
                force_options,
            )
        ).has_errors:
            raise LibraryError()
        resource.bundle.update(
            id_provider,
            bundle_element,
            container_options,
            network_options,
            port_map_add,
            port_map_remove,
            storage_map_add,
            storage_map_remove,
            meta_attributes,
        )


def _disable_validate_and_edit_cib(
    env: LibraryEnvironment, cib: _Element, resource_or_tag_ids: Iterable[str],
) -> List[_Element]:
    resource_el_list, report_list = _find_resources_expand_tags(
        cib, resource_or_tag_ids
    )
    env.report_processor.report_list(report_list)
    if env.report_processor.report_list(
        _resource_list_enable_disable(
            resource_el_list,
            resource.common.disable,
            IdProvider(cib),
            env.get_cluster_state(),
        )
    ).has_errors:
        raise LibraryError()
    return resource_el_list


def _disable_get_element_ids(
    disabled_resource_el_list: Iterable[_Element],
) -> Tuple[Set[str], Set[str]]:
    """
    Turn a list of elements asked by a user to be disabled to a list of their
    IDs and a list of IDs of their inner elements. Remember, the user can
    specify tags instead of resources. Therefore the list of disabled
    resources' IDs returned by this function may be different than the list of
    IDs entered in the command.
    """
    inner_resource_id_set = set()
    disabled_resource_id_set = set()
    for resource_el in disabled_resource_el_list:
        disabled_resource_id_set.add(cast(Optional[str], resource_el.get("id")))
        inner_resource_id_set.update(
            {
                cast(Optional[str], inner_resource_el.get("id"))
                for inner_resource_el in resource.common.get_all_inner_resources(
                    resource_el
                )
            }
        )
    # Make sure we only return found IDs and not None to match the function's
    # return type annotation.
    return (
        set(filter(None, disabled_resource_id_set)),
        set(filter(None, inner_resource_id_set)),
    )


def _disable_run_simulate(
    cmd_runner: CommandRunner,
    cib: _Element,
    disabled_resource_ids: Set[str],
    inner_resource_ids: Set[str],
    strict: bool,
) -> Tuple[str, Set[str]]:
    plaintext_status, transitions, dummy_cib = simulate_cib(cmd_runner, cib)
    simulated_operations = simulate_tools.get_operations_from_transitions(
        transitions
    )
    other_affected: Set[str] = set()
    if strict:
        other_affected = set(
            simulate_tools.get_resources_from_operations(
                simulated_operations, exclude=disabled_resource_ids
            )
        )
    else:
        other_affected = set(
            simulate_tools.get_resources_left_stopped(
                simulated_operations, exclude=disabled_resource_ids
            )
            + simulate_tools.get_resources_left_demoted(
                simulated_operations, exclude=disabled_resource_ids
            )
        )

    # Stopping a clone stops all its inner resources. That should not block
    # stopping the clone.
    other_affected = other_affected - inner_resource_ids
    return plaintext_status, other_affected


def disable(
    env: LibraryEnvironment,
    resource_or_tag_ids: Iterable[str],
    wait: WaitType = False,
):
    """
    Disallow specified resources to be started by the cluster

    env -- provides all for communication with externals
    resource_or_tag_ids -- ids of the resources to become disabled, or in case
        of tag ids, all resources in tags are to be disabled
    wait -- False: no wait, None: wait default timeout, int: wait timeout
    """
    env.ensure_wait_satisfiable(wait)
    _disable_validate_and_edit_cib(env, env.get_cib(), resource_or_tag_ids)
    _push_cib_wait(
        env, wait, resource_or_tag_ids, _ensure_disabled_after_wait(True)
    )


def disable_safe(
    env: LibraryEnvironment,
    resource_or_tag_ids: Iterable[str],
    strict: bool,
    wait: WaitType = False,
):
    """
    Disallow specified resources to be started by the cluster only if there is
    no effect on other resources

    env -- provides all for communication with externals
    resource_or_tag_ids -- ids of the resources to become disabled, or in case
        of tag ids, all resources in tags are to be disabled
    strict -- if False, allow resources to be migrated
    wait -- False: no wait, None: wait default timeout, int: wait timeout
    """
    # pylint: disable=too-many-locals
    if not env.is_cib_live:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentRequired([file_type_codes.CIB])
            )
        )

    env.ensure_wait_satisfiable(wait)
    cib = env.get_cib()
    resource_el_list = _disable_validate_and_edit_cib(
        env, cib, resource_or_tag_ids
    )
    disabled_resource_id_set, inner_resource_id_set = _disable_get_element_ids(
        resource_el_list
    )
    plaintext_status, other_affected = _disable_run_simulate(
        env.cmd_runner(),
        cib,
        disabled_resource_id_set,
        inner_resource_id_set,
        strict,
    )
    if other_affected:
        raise LibraryError(
            ReportItem.error(
                reports.messages.ResourceDisableAffectsOtherResources(
                    sorted(disabled_resource_id_set),
                    sorted(other_affected),
                    plaintext_status,
                )
            )
        )
    _push_cib_wait(
        env, wait, disabled_resource_id_set, _ensure_disabled_after_wait(True)
    )


def disable_simulate(
    env: LibraryEnvironment, resource_or_tag_ids: Iterable[str], strict: bool
) -> Mapping[str, Union[str, List[str]]]:
    """
    Simulate disallowing specified resources to be started by the cluster

    env -- provides all for communication with externals
    resource_or_tag_ids -- ids of the resources to become disabled, or in case
        of tag ids, all resources in tags are to be disabled
    bool strict -- if False, allow resources to be migrated
    """
    if not env.is_cib_live:
        raise LibraryError(
            ReportItem.error(
                reports.messages.LiveEnvironmentRequired([file_type_codes.CIB])
            )
        )

    cib = env.get_cib()
    resource_el_list = _disable_validate_and_edit_cib(
        env, cib, resource_or_tag_ids
    )
    disabled_resource_id_set, inner_resource_id_set = _disable_get_element_ids(
        resource_el_list
    )
    plaintext_status, other_affected = _disable_run_simulate(
        env.cmd_runner(),
        cib,
        disabled_resource_id_set,
        inner_resource_id_set,
        strict,
    )
    return dict(
        plaintext_simulated_status=plaintext_status,
        other_affected_resource_list=sorted(other_affected),
    )


def enable(
    env: LibraryEnvironment,
    resource_or_tag_ids: Iterable[str],
    wait: WaitType = False,
):
    """
    Allow specified resources to be started by the cluster

    env -- provides all for communication with externals
    resource_or_tag_ids -- ids of the resources to become enabled, or in case
        of tag ids, all resources in tags are to be enabled
    wait -- False: no wait, None: wait default timeout, int: wait timeout
    """
    env.ensure_wait_satisfiable(wait)
    cib = env.get_cib()
    resource_el_list, report_list = _find_resources_expand_tags(
        cib, resource_or_tag_ids
    )
    env.report_processor.report_list(report_list)

    to_enable_set = set()
    for el in resource_el_list:
        to_enable_set.update(resource.common.find_resources_to_enable(el))

    if env.report_processor.report_list(
        _resource_list_enable_disable(
            to_enable_set,
            resource.common.enable,
            IdProvider(cib),
            env.get_cluster_state(),
        )
    ).has_errors:
        raise LibraryError()
    _push_cib_wait(
        env,
        wait,
        [str(el.get("id", "")) for el in resource_el_list],
        _ensure_disabled_after_wait(False),
    )


def _resource_list_enable_disable(
    resource_el_list, func, id_provider, cluster_state
):
    report_list = []
    for resource_el in resource_el_list:
        res_id = resource_el.attrib["id"]
        try:
            if not is_resource_managed(cluster_state, res_id):
                report_list.append(
                    ReportItem.warning(
                        reports.messages.ResourceIsUnmanaged(res_id)
                    )
                )
            func(resource_el, id_provider)
        except ResourceNotFound:
            report_list.append(
                ReportItem.error(
                    reports.messages.IdNotFound(
                        res_id,
                        ["bundle", "clone", "group", "master", "primitive"],
                    )
                )
            )
    return report_list


def unmanage(
    env: LibraryEnvironment,
    resource_or_tag_ids: Iterable[str],
    with_monitor: bool = False,
) -> None:
    """
    Set specified resources not to be managed by the cluster

    env -- provides all for communication with externals
    resource_or_tag_ids -- ids of the resources to become unmanaged, or in case
        of tag ids, all resources in tags are to be managed
    with_monitor -- disable resources' monitor operations
    """
    cib = env.get_cib()
    resource_el_list, report_list = _find_resources_expand_tags(
        cib, resource_or_tag_ids, resource.common.find_resources_to_unmanage
    )
    env.report_processor.report_list(report_list)
    if env.report_processor.has_errors:
        raise LibraryError()

    primitives_set = set()
    for resource_el in resource_el_list:
        resource.common.unmanage(resource_el, IdProvider(cib))
        if with_monitor:
            primitives_set.update(resource.common.find_primitives(resource_el))

    for resource_el in primitives_set:
        for op in resource.operations.get_resource_operations(
            resource_el, ["monitor"]
        ):
            resource.operations.disable(op)
    env.push_cib()


def manage(
    env: LibraryEnvironment,
    resource_or_tag_ids: Iterable[str],
    with_monitor: bool = False,
) -> None:
    """
    Set specified resources to be managed by the cluster

    env -- provides all for communication with externals
    resource_or_tag_ids -- ids of the resources to become managed, or in case
        of tag id, all resources in tag are to be managed
    with_monitor -- enable resources' monitor operations
    """
    cib = env.get_cib()

    resource_el_list, report_list = _find_resources_expand_tags(
        cib, resource_or_tag_ids, resource.common.find_resources_to_manage
    )

    primitives_set = set()
    for resource_el in resource_el_list:
        resource.common.manage(resource_el, IdProvider(cib))
        primitives_set.update(resource.common.find_primitives(resource_el))

    for resource_el in sorted(
        primitives_set, key=lambda element: element.get("id", "")
    ):
        op_list = resource.operations.get_resource_operations(
            resource_el, ["monitor"]
        )
        if with_monitor:
            for op in op_list:
                resource.operations.enable(op)
        else:
            monitor_enabled = False
            for op in op_list:
                if resource.operations.is_enabled(op):
                    monitor_enabled = True
                    break
            if op_list and not monitor_enabled:
                # do not advise enabling monitors if there are none defined
                report_list.append(
                    ReportItem.warning(
                        reports.messages.ResourceManagedNoMonitorEnabled(
                            str(resource_el.get("id", ""))
                        )
                    )
                )
    if env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()
    env.push_cib()


def group_add(
    env,
    group_id,
    resource_id_list,
    adjacent_resource_id=None,
    put_after_adjacent=True,
    wait=False,
):
    """
    Move specified resources into an existing or new group

    LibraryEnvironment env provides all for communication with externals
    string group_id -- id of the target group
    iterable resource_id_list -- ids of resources to put into the group
    string adjacent_resource_id -- put resources beside this one if specified
    bool put_after_adjacent -- put resources after or before the adjacent one
    mixed wait -- flag for controlling waiting for pacemaker idle mechanism
    """
    with resource_environment(env, wait, [group_id]) as resources_section:
        id_provider = IdProvider(resources_section)

        validator = resource.hierarchy.ValidateMoveResourcesToGroupByIds(
            group_id,
            resource_id_list,
            adjacent_resource_id=adjacent_resource_id,
        )
        if env.report_processor.report_list(
            validator.validate(resources_section, id_provider)
        ).has_errors:
            raise LibraryError()

        # If we get no group element from the validator and there were no
        # errors, then the element does not exist and we can create it.
        group_element = validator.group_element()
        if group_element is None:
            group_element = resource.group.append_new(
                resources_section, group_id
            )

        resource.hierarchy.move_resources_to_group(
            group_element,
            validator.resource_element_list(),
            adjacent_resource=validator.adjacent_resource_element(),
            put_after_adjacent=put_after_adjacent,
        )


def get_failcounts(
    env, resource=None, node=None, operation=None, interval=None
):
    # pylint: disable=redefined-outer-name
    """
    List resources failcounts, optionally filtered by a resource, node or op

    LibraryEnvironment env
    string resource -- show failcounts for the specified resource only
    string node -- show failcounts for the specified node only
    string operation -- show failcounts for the specified operation only
    string interval -- show failcounts for the specified operation interval only
    """
    report_items = []
    if interval is not None and operation is None:
        report_items.append(
            ReportItem.error(
                reports.messages.PrerequisiteOptionIsMissing(
                    "interval", "operation"
                )
            )
        )
    if interval is not None:
        report_items.extend(
            ValueTimeInterval("interval").validate({"interval": interval})
        )
    if report_items:
        raise LibraryError(*report_items)

    interval_ms = (
        None if interval is None else timeout_to_seconds(interval) * 1000
    )

    all_failcounts = cib_status.get_resources_failcounts(
        get_status(env.get_cib())
    )
    return cib_status.filter_resources_failcounts(
        all_failcounts,
        resource=resource,
        node=node,
        operation=operation,
        interval=interval_ms,
    )


def move(env, resource_id, node=None, master=False, lifetime=None, wait=False):
    """
    Create a constraint to move a resource

    LibraryEnvironment env
    string resource_id -- id of a resource to be moved
    string node -- node to move the resource to, ban on the current node if None
    bool master -- limit the constraint to the Master role
    string lifetime -- lifespan of the constraint, forever if None
    mixed wait -- flag for controlling waiting for pacemaker idle mechanism
    """
    return _Move().run(
        env, resource_id, node=node, master=master, lifetime=lifetime, wait=wait
    )


def ban(env, resource_id, node=None, master=False, lifetime=None, wait=False):
    """
    Create a constraint to keep a resource of a node

    LibraryEnvironment env
    string resource_id -- id of a resource to be banned
    string node -- node to ban the resource on, ban on the current node if None
    bool master -- limit the constraint to the Master role
    string lifetime -- lifespan of the constraint, forever if None
    mixed wait -- flag for controlling waiting for pacemaker idle mechanism
    """
    return _Ban().run(
        env, resource_id, node=node, master=master, lifetime=lifetime, wait=wait
    )


class _MoveBanTemplate:
    def _validate(self, resource_el, master):
        raise NotImplementedError()

    def _run_action(self, runner, resource_id, node, master, lifetime):
        raise NotImplementedError()

    def _report_action_stopped_resource(self, resource_id):
        raise NotImplementedError()

    def _report_action_pcmk_error(self, resource_id, stdout, stderr):
        raise NotImplementedError()

    def _report_action_pcmk_success(self, resource_id, stdout, stderr):
        raise NotImplementedError()

    def _report_wait_result(
        self,
        resource_id,
        node,
        resource_running_on_before,
        resource_running_on_after,
    ):
        raise NotImplementedError()

    @staticmethod
    def _running_on_nodes(resource_state):
        if resource_state:
            return frozenset(
                resource_state.get("Master", [])
                + resource_state.get("Started", [])
            )
        return frozenset()

    def run(
        self,
        env,
        resource_id,
        node=None,
        master=False,
        lifetime=None,
        wait=False,
    ):
        # validate
        env.ensure_wait_satisfiable(wait)  # raises on error

        resource_el, report_list = resource.common.find_one_resource(
            get_resources(env.get_cib()), resource_id
        )
        if resource_el is not None:
            report_list.extend(self._validate(resource_el, master))
        if env.report_processor.report_list(report_list).has_errors:
            raise LibraryError()

        # get current status for wait processing
        if wait is not False:
            resource_running_on_before = get_resource_state(
                env.get_cluster_state(), resource_id
            )

        # run the action
        stdout, stderr, retval = self._run_action(
            env.cmd_runner(),
            resource_id,
            node=node,
            master=master,
            lifetime=lifetime,
        )
        if retval != 0:
            if (
                f"Resource '{resource_id}' not moved: active in 0 locations"
                in stderr
            ):
                raise LibraryError(
                    self._report_action_stopped_resource(resource_id)
                )
            raise LibraryError(
                self._report_action_pcmk_error(resource_id, stdout, stderr)
            )
        env.report_processor.report(
            self._report_action_pcmk_success(resource_id, stdout, stderr)
        )

        # process wait
        if wait is not False:
            wait_for_idle(env.cmd_runner(), env.get_wait_timeout(wait))
            resource_running_on_after = get_resource_state(
                env.get_cluster_state(), resource_id
            )
            if env.report_processor.report(
                self._report_wait_result(
                    resource_id,
                    node,
                    resource_running_on_before,
                    resource_running_on_after,
                )
            ).has_errors:
                raise LibraryError()


class _Move(_MoveBanTemplate):
    def _validate(self, resource_el, master):
        return resource.common.validate_move(resource_el, master)

    def _run_action(self, runner, resource_id, node, master, lifetime):
        return resource_move(
            runner, resource_id, node=node, master=master, lifetime=lifetime
        )

    def _report_action_stopped_resource(self, resource_id):
        return ReportItem.error(
            reports.messages.CannotMoveResourceStoppedNoNodeSpecified(
                resource_id
            )
        )

    def _report_action_pcmk_error(self, resource_id, stdout, stderr):
        return ReportItem.error(
            reports.messages.ResourceMovePcmkError(resource_id, stdout, stderr)
        )

    def _report_action_pcmk_success(self, resource_id, stdout, stderr):
        return ReportItem.info(
            reports.messages.ResourceMovePcmkSuccess(
                resource_id, stdout, stderr,
            )
        )

    def _report_wait_result(
        self,
        resource_id,
        node,
        resource_running_on_before,
        resource_running_on_after,
    ):
        allowed_nodes = frozenset([node] if node else [])
        running_on_nodes = self._running_on_nodes(resource_running_on_after)

        severity = reports.item.ReportItemSeverity.info()
        if resource_running_on_before and (  # running resource moved
            not running_on_nodes
            or (allowed_nodes and allowed_nodes.isdisjoint(running_on_nodes))
            or (resource_running_on_before == resource_running_on_after)
        ):
            severity = reports.item.ReportItemSeverity.error()
        if not resource_running_on_after:
            return ReportItem(
                severity, reports.messages.ResourceDoesNotRun(resource_id),
            )
        return ReportItem(
            severity,
            reports.messages.ResourceRunningOnNodes(
                resource_id, resource_running_on_after,
            ),
        )


class _Ban(_MoveBanTemplate):
    def _validate(self, resource_el, master):
        return resource.common.validate_ban(resource_el, master)

    def _run_action(self, runner, resource_id, node, master, lifetime):
        return resource_ban(
            runner, resource_id, node=node, master=master, lifetime=lifetime
        )

    def _report_action_stopped_resource(self, resource_id):
        return ReportItem.error(
            reports.messages.CannotBanResourceStoppedNoNodeSpecified(
                resource_id,
            )
        )

    def _report_action_pcmk_error(self, resource_id, stdout, stderr):
        return ReportItem.error(
            reports.messages.ResourceBanPcmkError(resource_id, stdout, stderr)
        )

    def _report_action_pcmk_success(
        self, resource_id: str, stdout: str, stderr: str,
    ) -> ReportItem:
        return ReportItem.info(
            reports.messages.ResourceBanPcmkSuccess(resource_id, stdout, stderr)
        )

    def _report_wait_result(
        self,
        resource_id,
        node,
        resource_running_on_before,
        resource_running_on_after,
    ):
        running_on_nodes = self._running_on_nodes(resource_running_on_after)
        if node:
            banned_nodes = frozenset([node])
        else:
            banned_nodes = self._running_on_nodes(resource_running_on_before)

        severity = reports.item.ReportItemSeverity.info()
        if not banned_nodes.isdisjoint(running_on_nodes) or (
            resource_running_on_before and not running_on_nodes
        ):
            severity = reports.item.ReportItemSeverity.error()
        if not resource_running_on_after:
            return ReportItem(
                severity, reports.messages.ResourceDoesNotRun(resource_id),
            )
        return ReportItem(
            severity,
            reports.messages.ResourceRunningOnNodes(
                resource_id, resource_running_on_after,
            ),
        )


def unmove_unban(
    env, resource_id, node=None, master=False, expired=False, wait=False
):
    """
    Remove all constraints created by move and ban

    LibraryEnvironment env
    string resource_id -- id of a resource to be unmoved/unbanned
    string node -- node to limit unmoving/unbanning to, all nodes if None
    bool master -- only remove constraints for Master role
    bool expired -- only remove constrains which have already expired
    mixed wait -- flag for controlling waiting for pacemaker idle mechanism
    """
    # validate
    env.ensure_wait_satisfiable(wait)  # raises on error

    resource_el, report_list = resource.common.find_one_resource(
        get_resources(env.get_cib()), resource_id
    )
    if resource_el is not None:
        report_list.extend(
            resource.common.validate_unmove_unban(resource_el, master)
        )
    if expired and not has_resource_unmove_unban_expired_support(
        env.cmd_runner()
    ):
        report_list.append(
            ReportItem.error(
                reports.messages.ResourceUnmoveUnbanPcmkExpiredNotSupported()
            )
        )
    if env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()

    # run the action
    stdout, stderr, retval = resource_unmove_unban(
        env.cmd_runner(), resource_id, node=node, master=master, expired=expired
    )
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.ResourceUnmoveUnbanPcmkError(
                    resource_id, stdout, stderr
                )
            )
        )
    env.report_processor.report(
        ReportItem.info(
            reports.messages.ResourceUnmoveUnbanPcmkSuccess(
                resource_id, stdout, stderr
            )
        )
    )

    # process wait
    if wait is not False:
        wait_for_idle(env.cmd_runner(), env.get_wait_timeout(wait))
        if env.report_processor.report(
            info_resource_state(env.get_cluster_state(), resource_id)
        ).has_errors:
            raise LibraryError()


def get_resource_relations_tree(
    env: LibraryEnvironment, resource_id: str,
) -> Mapping[str, Any]:
    """
    Return a dict representing tree-like structure of resources and their
    relations.

    env -- library environment
    resource_id -- id of a resource which should be the root of the relation
        tree
    """
    cib = env.get_cib()

    dummy_resource_el, report_list = resource.common.find_one_resource(
        get_resources(cib), resource_id
    )
    if env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()

    (
        resources_dict,
        relations_dict,
    ) = resource.relations.ResourceRelationsFetcher(cib).get_relations(
        resource_id
    )
    return dto.to_dict(
        resource.relations.ResourceRelationTreeBuilder(
            resources_dict, relations_dict
        )
        .get_tree(resource_id)
        .to_dto()
    )


def _find_resources_expand_tags(
    cib: _Element,
    resource_or_tag_ids: Iterable[str],
    additional_search: Optional[Callable[[_Element], List[_Element]]] = None,
) -> Tuple[List[_Element], ReportItemList]:
    rsc_or_tag_el_list, report_list = resource.common.find_resources(
        cib,
        resource_or_tag_ids,
        resource_tags=resource.common.ALL_RESOURCE_XML_TAGS + [TAG_TAG],
    )

    resource_set = set()
    for el in rsc_or_tag_el_list:
        resource_set.update(
            expand_tag(
                el, only_expand_types=resource.common.ALL_RESOURCE_XML_TAGS
            )
        )
    if not additional_search:
        return list(resource_set), report_list

    final_set = set()
    for el in resource_set:
        final_set.update(additional_search(el))
    return list(final_set), report_list


def get_required_cib_version_for_primitive(
    op_list: Iterable[Mapping[str, str]]
) -> Optional[Version]:
    for op in op_list:
        if op.get("on-fail", "") == "demote":
            return Version(3, 4, 0)
    return None



from functools import partial

from lxml import etree

from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.lib.cib.constraint import constraint
from pcs.lib.cib import tools
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import RESOURCE_ROLES
from pcs.lib.xml_tools import remove_when_pointless

TAG_NAME = "rsc_ticket"
DESCRIPTION = "constraint id"
ATTRIB = {
    "loss-policy": ("fence", "stop", "freeze", "demote"),
    "ticket": None,
}
ATTRIB_PLAIN = {
    "rsc": None,
    "rsc-role": RESOURCE_ROLES,
}


def _validate_options_common(options):
    report_list = []
    if "loss-policy" in options:
        loss_policy = options["loss-policy"].lower()
        if options["loss-policy"] not in ATTRIB["loss-policy"]:
            report_list.append(
                ReportItem.error(
                    reports.messages.InvalidOptionValue(
                        "loss-policy",
                        options["loss-policy"],
                        ATTRIB["loss-policy"],
                    )
                )
            )
        options["loss-policy"] = loss_policy
    return report_list


def _create_id(cib, ticket, resource_id, resource_role):
    return tools.find_unique_id(
        cib,
        "-".join(("ticket", ticket, resource_id))
        + ("-{0}".format(resource_role) if resource_role else ""),
    )


def prepare_options_with_set(cib, options, resource_set_list):
    options = constraint.prepare_options(
        tuple(ATTRIB.keys()),
        options,
        create_id_fn=partial(
            constraint.create_id, cib, "ticket", resource_set_list
        ),
        validate_id=partial(tools.check_new_id_applicable, cib, DESCRIPTION),
    )
    report_list = _validate_options_common(options)
    if "ticket" not in options or not options["ticket"].strip():
        report_list.append(
            ReportItem.error(
                reports.messages.RequiredOptionsAreMissing(["ticket"])
            )
        )
    if report_list:
        raise LibraryError(*report_list)
    return options


def prepare_options_plain(cib, options, ticket, resource_id):
    options = options.copy()

    report_list = _validate_options_common(options)

    if not ticket:
        report_list.append(
            ReportItem.error(
                reports.messages.RequiredOptionsAreMissing(["ticket"])
            )
        )
    options["ticket"] = ticket

    if not resource_id:
        report_list.append(
            ReportItem.error(
                reports.messages.RequiredOptionsAreMissing(["rsc"])
            )
        )
    options["rsc"] = resource_id

    if "rsc-role" in options:
        if options["rsc-role"]:
            resource_role = options["rsc-role"].lower().capitalize()
            if resource_role not in ATTRIB_PLAIN["rsc-role"]:
                report_list.append(
                    ReportItem.error(
                        reports.messages.InvalidOptionValue(
                            "rsc-role",
                            options["rsc-role"],
                            ATTRIB_PLAIN["rsc-role"],
                        )
                    )
                )
            options["rsc-role"] = resource_role
        else:
            del options["rsc-role"]

    if report_list:
        raise LibraryError(*report_list)

    return constraint.prepare_options(
        tuple(list(ATTRIB) + list(ATTRIB_PLAIN)),
        options,
        partial(
            _create_id,
            cib,
            options["ticket"],
            resource_id,
            options.get("rsc-role", ""),
        ),
        partial(tools.check_new_id_applicable, cib, DESCRIPTION),
    )


def create_plain(constraint_section, options):
    element = etree.SubElement(constraint_section, TAG_NAME)
    element.attrib.update(options)
    return element


def remove_plain(constraint_section, ticket_key, resource_id):
    ticket_element_list = constraint_section.xpath(
        './/rsc_ticket[@ticket="{0}" and @rsc="{1}"]'.format(
            ticket_key, resource_id
        )
    )

    for ticket_element in ticket_element_list:
        ticket_element.getparent().remove(ticket_element)

    return len(ticket_element_list) > 0


def remove_with_resource_set(constraint_section, ticket_key, resource_id):
    ref_element_list = constraint_section.xpath(
        (
            './/rsc_ticket[@ticket="{0}"]/resource_set/resource_ref[@id="{1}"]'
        ).format(ticket_key, resource_id)
    )

    for ref_element in ref_element_list:
        set_element = ref_element.getparent()
        set_element.remove(ref_element)
        # set_element is lxml element, therefore we have to use len() here
        # pylint: disable=len-as-condition
        if not len(set_element):
            ticket_element = set_element.getparent()
            ticket_element.remove(set_element)
            # We do not care about attributes since without an attribute "rsc"
            # they are pointless. Attribute "rsc" is mutually exclusive with
            # resource_set (see rng) so it cannot be in this ticket_element.
            remove_when_pointless(ticket_element, attribs_important=False)

    return len(ref_element_list) > 0


def are_duplicate_plain(element, other_element):
    return all(
        element.attrib.get(name, "") == other_element.attrib.get(name, "")
        for name in ("ticket", "rsc", "rsc-role")
    )


def are_duplicate_with_resource_set(element, other_element):
    return element.attrib["ticket"] == other_element.attrib[
        "ticket"
    ] and constraint.have_duplicate_resource_sets(element, other_element)



from unittest import TestCase

from pcs_test.tools.misc import outdent


class OutdentTest(TestCase):
    def test_returns_the_same_text_when_not_indented(self):
        text = "\n".join(["first line", "  second line", "    third line",])
        self.assertEqual(text, outdent(text))

    def test_remove_the_smallest_indentation(self):
        self.assertEqual(
            "\n".join(["  first line", "second line", "  third line",]),
            outdent(
                "\n".join(
                    ["    first line", "  second line", "    third line",]
                )
            ),
        )

    def test_very_ugly_indented_text(self):
        self.assertEqual(
            """\
Cluster Name: test99
  Options:
""",
            outdent(
                """\
                Cluster Name: test99
                  Options:
                """
            ),
        )



from textwrap import dedent

from pcs_test.tools.command_env.mock_runner import Call as RunnerCall

from pcs import settings


class SystemctlShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def disable(
        self,
        service,
        name="runner.systemctl.disable",
        stdout="",
        stderr="",
        returncode=0,
    ):
        self.__calls.place(
            name,
            RunnerCall(
                [settings.systemctl_binary, "disable", f"{service}.service"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def enable(
        self,
        service,
        name="runner.systemctl.enable",
        stdout="",
        stderr="",
        returncode=0,
    ):
        self.__calls.place(
            name,
            RunnerCall(
                [settings.systemctl_binary, "enable", f"{service}.service"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def start(
        self,
        service,
        name="runner.systemctl.start",
        stdout="",
        stderr="",
        returncode=0,
    ):
        self.__calls.place(
            name,
            RunnerCall(
                [settings.systemctl_binary, "start", f"{service}.service"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def stop(
        self,
        service,
        name="runner.systemctl.stop",
        stdout="",
        stderr="",
        returncode=0,
    ):
        self.__calls.place(
            name,
            RunnerCall(
                [settings.systemctl_binary, "stop", f"{service}.service"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def is_active(
        self, service, name="runner_systemctl.is_active", is_active=True
    ):
        args = dict(stdout="unknown\n", returncode=3,)
        if is_active:
            args = dict(stdout="active\n", returncode=0,)
        self.__calls.place(
            name,
            RunnerCall(
                [settings.systemctl_binary, "is-active", f"{service}.service"],
                **args,
            ),
        )

    def is_enabled(
        self,
        service,
        is_enabled=True,
        name="runner_systemctl.is_enabled",
        before=None,
        instead=None,
    ):
        args = dict(stdout="disabled\n", returncode=1,)
        if is_enabled:
            args = dict(stdout="enabled\n", returncode=0,)
        self.__calls.place(
            name,
            RunnerCall(
                [settings.systemctl_binary, "is-enabled", f"{service}.service"],
                **args,
            ),
            before=before,
            instead=instead,
        )

    def list_unit_files(
        self,
        unit_file_states,
        name="runner_systemctl.list_unit_files",
        before=None,
        instead=None,
    ):
        if not unit_file_states:
            output = dedent(
                """\
                UNIT FILE   STATE

                0 unit files listed.
                """
            )
        else:
            unit_len = max(len(x) for x in unit_file_states) + len(".service")
            state_len = max(len(x) for x in unit_file_states.values())
            pattern = "{unit:<{unit_len}} {state:<{state_len}}"
            lines = (
                [
                    pattern.format(
                        unit="UNIT FILE",
                        unit_len=unit_len,
                        state="STATE",
                        state_len=state_len,
                    )
                ]
                + [
                    pattern.format(
                        unit="{0}.service".format(unit),
                        unit_len=unit_len,
                        state=state,
                        state_len=state_len,
                    )
                    for unit, state in unit_file_states.items()
                ]
                + ["", "{0} unit files listed.".format(len(unit_file_states))]
            )
            output = "\n".join(lines)

        self.__calls.place(
            name,
            RunnerCall(
                [settings.systemctl_binary, "list-unit-files", "--full"],
                stdout=output,
            ),
            before=before,
            instead=instead,
        )



import os
from unittest import mock, TestCase

from OpenSSL import SSL

from pcs_test.tools.misc import get_tmp_dir

from pcs.daemon.ssl import PcsdSSL, CertKeyPair, SSLCertKeyException

SERVER_NAME = "pcsd-daemon"
SSL_OPTIONS = 0
SSL_CIPHERS = "DEFAULT:!RC4"


class SslFilesMixin:
    def setUp(self):
        # pylint cannot possibly know this is being mixed into TestCase classes
        # pylint: disable=invalid-name
        self.ssl_dir = get_tmp_dir("tier0_daemon_ssl")
        self.cert_path = os.path.join(self.ssl_dir.name, "daemon.cert")
        self.key_path = os.path.join(self.ssl_dir.name, "daemon.key")
        # various versions of OpenSSL / PyOpenSSL emit different messages
        self.DAMAGED_SSL_FILES_ERRORS_1 = (
            f"Invalid SSL certificate '{self.cert_path}':"
            " 'PEM routines:PEM_read_bio:no start line'",
            f"Invalid SSL key '{self.key_path}':"
            " 'PEM routines:PEM_read_bio:no start line'",
        )
        self.DAMAGED_SSL_FILES_ERRORS_2 = (
            f"Invalid SSL certificate '{self.cert_path}':"
            " 'PEM routines:get_name:no start line'",
            f"Invalid SSL key '{self.key_path}':"
            " 'PEM routines:get_name:no start line'",
        )

    def tearDown(self):
        # pylint cannot possibly know this is being mixed into TestCase classes
        # pylint: disable=invalid-name
        self.ssl_dir.cleanup()

    def damage_ssl_files(self):
        with open(self.cert_path, "w") as cert:
            cert.write("bad content")
        with open(self.key_path, "w") as key:
            key.write("bad content")


class Pair(SslFilesMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.pair = CertKeyPair(self.cert_path, self.key_path)

    def test_generated_key_check_success(self):
        self.pair.regenerate(SERVER_NAME)
        self.assertEqual(self.pair.check(), [])

    def test_error_if_files_with_bad_content(self):
        self.damage_ssl_files()
        self.assertTrue(
            self.pair.check()
            in [
                list(self.DAMAGED_SSL_FILES_ERRORS_1),
                list(self.DAMAGED_SSL_FILES_ERRORS_2),
            ]
        )

    @mock.patch("pcs.daemon.ssl.SSL.Context.use_privatekey")
    def test_error_if_short_key(self, mock_use_key):
        mock_use_key.side_effect = SSL.Error("reason")
        self.pair.regenerate(SERVER_NAME, 1024)
        errors = self.pair.check()
        self.assertEqual(
            errors, ["Unable to load SSL certificate and/or key: reason"]
        )

    def test_error_if_cert_does_not_match_key(self):
        self.pair.regenerate(SERVER_NAME)
        with open(self.key_path, "rb") as key_file:
            key_content = key_file.read()
        self.pair.regenerate(SERVER_NAME)
        with open(self.key_path, "wb") as key_file:
            key_file.write(key_content)

        errors = self.pair.check()
        self.assertEqual(len(errors), 1)
        self.assertTrue(
            errors[0].startswith("SSL certificate does not match the key:")
        )


class PcsdSSLTest(SslFilesMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.pcsd_ssl = PcsdSSL(
            SERVER_NAME, self.cert_path, self.key_path, SSL_OPTIONS, SSL_CIPHERS
        )

    def test_is_ok_when_ssl_files_does_not_exist(self):
        self.pcsd_ssl.guarantee_valid_certs()

    def test_raises_when_ssl_files_are_damaged(self):
        self.damage_ssl_files()
        with self.assertRaises(SSLCertKeyException) as ctx_manager:
            self.pcsd_ssl.guarantee_valid_certs()
        self.assertTrue(
            ctx_manager.exception.args
            in [
                self.DAMAGED_SSL_FILES_ERRORS_1,
                self.DAMAGED_SSL_FILES_ERRORS_2,
            ]
        )

    def test_context_uses_given_options(self):
        self.pcsd_ssl.guarantee_valid_certs()
        ssl_context = self.pcsd_ssl.create_context()
        self.assertEqual(ssl_context.options, SSL_OPTIONS)
